/*
 * This module handles management transactions requested by privileged
 * manage programs.  When the remote port and address of a connection is
 * the management port and address, the listener calls this routine rather
 * than create a normal client thread.
 *
 * Requests:
 *	SHUTDOWN[/grace-period]
 *	RESTART[/grace-period]
 *	INVCACHE
 *
 * The http_manage_request function is called directly by the port listener 
 * thread created by ts_declare_tcp_port() and therefore has the following
 * considerations:
 *
 *	> Context is not fully initialized, only safe actions are ts_tcp_read
 *	  and ts_tcp_write.
 *
 *	> New connections on the receiving port are suspended until this
 *	  routine returns.  Note that since more than one listener thread is
 *	  allowed, multiple active management requests are still possible.
 *
 *	> Stacksize is much smaller than for a normal client port, limiting
 *	  max request size
 *
 *	> Scheduler policy is FIFO, rather than DEFAULT, giving this routine
 *	  priority over client threads.
 *
 *  Author:	David Jones
 *  Date:	20-AUG-1994
 */
#include <stdlib.h>
#include <stdio.h>
#include "tutil.h"
int ts_tcp_read(), ts_tcp_write(), tlog_putlog();

int http_manage_request ( void *ctx,	/* I/O context */
	short port,			/* Local port number */
	int *shutdown_time )
{
    int status, length;
    char *stsmsg, cmd[100];
    struct tu_streambuf stream;
    /*
     * Initialize stream and read command (delimited by newline).
     */
    tu_init_stream ( ctx, ts_tcp_read, &stream );

    status = tu_read_line ( &stream, cmd, sizeof(cmd)-1, &length );
    if ( (status&1) == 0 ) return 0;	/* ignore errors */
    /*
     * Upcase command string and process, set status to value to be returned.
     */
    status = 0;
    *shutdown_time = 10;
    cmd[length] = '\0';
    tlog_putlog ( 1, "Management request to port !SW: '!AZ'!/", port, cmd );
    tu_strupcase ( cmd, cmd );
    if ( 0 == tu_strncmp(cmd,"SHUTDOWN",8) ) {
	/*
	 * Set return status to value that causes port to shutdown,
	 * waiting at most shutdown_time seconds.
	 */
	stsmsg = "201 Shutting down server port, active: ";
	status = 1;

    } else if ( 0 == tu_strncmp(cmd,"RESTART",7) ) {
	/*
	 * For restart, set status to value that causes port listener to
	 * rundown with status that causes server restart.
	 */
	stsmsg = "201 Restarting server active clients: ";
        status = 3;

    } else if ( 0 == tu_strncmp(cmd,"INVCACHE", 9) ) {
	int http_invalidate_doc_cache();
	http_invalidate_doc_cache();	/* clear cache */
	stsmsg = "200 Cache marked invalid\r\n";

    } else {
	stsmsg = "400 Management request invalid\r\n";
    }
    ts_tcp_write ( ctx, stsmsg, tu_strlen ( stsmsg ) );
    return status;
}
